@@ -28,21 +28,23 @@
 #include <basewrapper.h>
 #include <bindingmanager.h>
 #include <gilstate.h>
+#include <helper.h>
 #include <sbkconverter.h>
 #include <sbkstring.h>
 #include <sbkstaticstrings.h>
 #include <sbkfeature_base.h>
 #include <sbkmodule.h>
 
-#include <QtCore/QByteArray>
-#include <QtCore/QCoreApplication>
-#include <QtCore/QDebug>
-#include <QtCore/QDir>
-#include <QtCore/QFileInfo>
-#include <QtCore/QMetaMethod>
-#include <QtCore/QMutex>
-#include <QtCore/QStack>
-#include <QtCore/QThread>
+#include <QtCore/qbytearray.h>
+#include <QtCore/qcoreapplication.h>
+#include <QtCore/qdebug.h>
+#include <QtCore/qdir.h>
+#include <QtCore/qfileinfo.h>
+#include <QtCore/qmetaobject.h>
+#include <QtCore/qmutex.h>
+#include <QtCore/qstack.h>
+#include <QtCore/qthread.h>
+#include <QtCore/private/qobject_p.h>
 
 #include <algorithm>
 #include <cstring>
@@ -51,6 +53,13 @@
 #include <optional>
 #include <typeinfo>
 
+#ifdef Q_OS_WIN
+#  include <conio.h>
+#else
+#  include <QtCore/qdeadlinetimer.h>
+#  include <QtCore/private/qcore_unix_p.h>
+#endif
+
 using namespace Qt::StringLiterals;
 
 static QStack<PySide::CleanupFunction> cleanupFunctionList;
@@ -63,6 +72,20 @@ QT_END_NAMESPACE
 
 Q_LOGGING_CATEGORY(lcPySide, "qt.pyside.libpyside", QtCriticalMsg)
 
+static QObjectData *qt_object_private(const QObject *o)
+{
+    class FriendlyQObject : public QObject {
+    public:
+        using QObject::d_ptr;
+    };
+    return static_cast<const FriendlyQObject *>(o)->d_ptr.data();
+}
+
+static bool hasDynamicMetaObject(const QObject *o)
+{
+    return qt_object_private(o)->metaObject != nullptr;
+}
+
 namespace PySide
 {
 
@@ -76,7 +99,7 @@ void init(PyObject *module)
     ClassProperty::init(module);
     MetaFunction::init(module);
     // Init signal manager, so it will register some meta types used by QVariant.
-    SignalManager::instance();
+    SignalManager::init();
     initQApp();
 }
 
@@ -200,13 +223,13 @@ static QByteArrayList _SbkType_LookupProperty(PyTypeObject *type,
     if (origName.isEmpty())
         return QByteArrayList{};
     PyObject *mro = type->tp_mro;
-    auto n = PyTuple_GET_SIZE(mro);
+    auto n = PyTuple_Size(mro);
     auto len = std::strlen(origName);
     for (Py_ssize_t idx = 0; idx < n; idx++) {
-        PyTypeObject *base = reinterpret_cast<PyTypeObject *>(PyTuple_GET_ITEM(mro, idx));
+        auto *base = reinterpret_cast<PyTypeObject *>(PyTuple_GetItem(mro, idx));
         if (!SbkObjectType_Check(base))
             continue;
-        auto props = SbkObjectType_GetPropertyStrings(base);
+        auto *props = SbkObjectType_GetPropertyStrings(base);
         if (props == nullptr || *props == nullptr)
             continue;
         for (; *props != nullptr; ++props) {
@@ -240,11 +263,12 @@ static bool _setProperty(PyObject *qObj, PyObject *name, PyObject *value, bool *
     using Shiboken::AutoDecRef;
 
     QByteArray propName(Shiboken::String::toCString(name));
-    auto type = Py_TYPE(qObj);
+    auto *type = Py_TYPE(qObj);
     int flags = currentSelectId(type);
     int prop_flag = flags & 0x02;
     auto found = false;
-    QByteArray getterName{}, setterName{};
+    QByteArray getterName;
+    QByteArray setterName;
 
     auto fields = _SbkType_LookupProperty(type, propName, flags);
     if (!fields.isEmpty()) {
@@ -316,7 +340,7 @@ static std::optional<QMetaMethod> findSignal(const QMetaObject *mo,
         if (method.methodType() == QMetaMethod::Signal && method.name() == name)
             return method;
     }
-    auto *base = mo->superClass();
+    const auto *base = mo->superClass();
     return base != nullptr ? findSignal(base, name) : std::nullopt;
 }
 
@@ -324,7 +348,8 @@ bool fillQtProperties(PyObject *qObj, const QMetaObject *metaObj,
                       PyObject *kwds, bool allowErrors)
 {
 
-    PyObject *key, *value;
+    PyObject *key{};
+    PyObject *value{};
     Py_ssize_t pos = 0;
     int flags = currentSelectId(Py_TYPE(qObj));
     int snake_flag = flags & 0x01;
@@ -381,9 +406,9 @@ void runCleanupFunctions()
 
 static void destructionVisitor(SbkObject *pyObj, void *data)
 {
-    auto realData = reinterpret_cast<void **>(data);
-    auto pyQApp = reinterpret_cast<SbkObject *>(realData[0]);
-    auto pyQObjectType = reinterpret_cast<PyTypeObject *>(realData[1]);
+    auto *realData = reinterpret_cast<void **>(data);
+    auto *pyQApp = reinterpret_cast<SbkObject *>(realData[0]);
+    auto *pyQObjectType = reinterpret_cast<PyTypeObject *>(realData[1]);
 
     if (pyObj != pyQApp && PyObject_TypeCheck(pyObj, pyQObjectType)) {
         if (Shiboken::Object::hasOwnership(pyObj) && Shiboken::Object::isValid(pyObj, false)) {
@@ -402,7 +427,6 @@ void destroyQCoreApplication()
     QCoreApplication *app = QCoreApplication::instance();
     if (!app)
         return;
-    SignalManager::instance().clear();
 
     Shiboken::BindingManager &bm = Shiboken::BindingManager::instance();
     SbkObject *pyQApp = bm.retrieveWrapper(app);
@@ -431,7 +455,7 @@ std::size_t getSizeOfQObject(PyTypeObject *type)
 void initDynamicMetaObject(PyTypeObject *type, const QMetaObject *base, std::size_t cppObjSize)
 {
     //create DynamicMetaObject based on python type
-    auto userData = new TypeUserData(reinterpret_cast<PyTypeObject *>(type), base, cppObjSize);
+    auto *userData = new TypeUserData(reinterpret_cast<PyTypeObject *>(type), base, cppObjSize);
     userData->mo.update();
     Shiboken::ObjectType::setTypeUserData(type, userData, Shiboken::callCppDestructor<TypeUserData>);
 
@@ -454,7 +478,7 @@ TypeUserData *retrieveTypeUserData(PyTypeObject *pyTypeObj)
 
 TypeUserData *retrieveTypeUserData(PyObject *pyObj)
 {
-    auto pyTypeObj = PyType_Check(pyObj)
+    auto *pyTypeObj = PyType_Check(pyObj)
         ? reinterpret_cast<PyTypeObject *>(pyObj) : Py_TYPE(pyObj);
     return retrieveTypeUserData(pyTypeObj);
 }
@@ -467,7 +491,7 @@ const QMetaObject *retrieveMetaObject(PyTypeObject *pyTypeObj)
 
 const QMetaObject *retrieveMetaObject(PyObject *pyObj)
 {
-    auto pyTypeObj = PyType_Check(pyObj)
+    auto *pyTypeObj = PyType_Check(pyObj)
         ? reinterpret_cast<PyTypeObject *>(pyObj) : Py_TYPE(pyObj);
     return retrieveMetaObject(pyTypeObj);
 }
@@ -476,20 +500,19 @@ void initQObjectSubType(PyTypeObject *type, PyObject *args, PyObject * /* kwds *
 {
     PyTypeObject *qObjType = Shiboken::Conversions::getPythonTypeObject("QObject*");
 
-    PyObject *bases = PyTuple_GET_ITEM(args, 1);
-    int numBases = PyTuple_GET_SIZE(bases);
+    PyObject *bases = PyTuple_GetItem(args, 1);
 
     TypeUserData *userData = nullptr;
 
-    for (int i = 0; i < numBases; ++i) {
-        auto base = reinterpret_cast<PyTypeObject *>(PyTuple_GET_ITEM(bases, i));
+    for (Py_ssize_t i = 0, numBases = PyTuple_Size(bases); i < numBases; ++i) {
+        auto *base = reinterpret_cast<PyTypeObject *>(PyTuple_GetItem(bases, i));
         if (PyType_IsSubtype(base, qObjType)) {
             userData = retrieveTypeUserData(base);
             break;
         }
     }
     if (!userData) {
-        const char *className = Shiboken::String::toCString(PyTuple_GET_ITEM(args, 0));
+        const char *className = Shiboken::String::toCString(PyTuple_GetItem(args, 0));
         qWarning("Sub class of QObject not inheriting QObject!? Crash will happen when using %s.",
                  className);
         return;
@@ -501,6 +524,36 @@ void initQObjectSubType(PyTypeObject *type, PyObject *args, PyObject * /* kwds *
     PySide::Feature::Enable(true);
 }
 
+extern "C" {
+static int qAppInputHook()
+{
+    auto *app = QCoreApplication::instance();
+    if (app == nullptr || app->thread() != QThread::currentThread())
+        return 0;
+#ifndef Q_OS_WIN
+    // Check for press on stdin (file descriptor 0)
+    pollfd stdinPfd =  qt_make_pollfd(0, POLLIN);
+    while (qt_safe_poll(&stdinPfd, 1, QDeadlineTimer{1}) == 0)
+        QCoreApplication::processEvents({}, 50000);
+#else
+    while (_kbhit() == 0)
+        QCoreApplication::processEvents({}, 50000);
+#endif
+    return 0;
+}
+} // extern "C"
+
+static void unregisterQAppInputHook()
+{
+    PyOS_InputHook = nullptr;
+}
+
+static void registerQAppInputHook()
+{
+    PyOS_InputHook = qAppInputHook;
+    qAddPostRoutine(unregisterQAppInputHook);
+}
+
 void initQApp()
 {
     /*
@@ -514,8 +567,10 @@ void initQApp()
      * I would appreciate very much if someone could explain or even fix
      * this issue. It exists only when a pre-existing application exists.
      */
-    if (!qApp)
+    if (qApp == nullptr) {
+        registerQAppInputHook();
         Py_DECREF(MakeQAppWrapper(nullptr));
+    }
 
     // PYSIDE-1470: Register a function to destroy an application from shiboken.
     setDestroyQApplication(destroyQCoreApplication);
@@ -540,7 +595,9 @@ PyObject *getHiddenDataFromQObject(QObject *cppSelf, PyObject *self, PyObject *n
 
     // Search on metaobject (avoid internal attributes started with '__')
     if (!attr) {
-        PyObject *type, *value, *traceback;
+        PyObject *type{};
+        PyObject *value{};
+        PyObject *traceback{};
         PyErr_Fetch(&type, &value, &traceback);     // This was omitted for a loong time.
 
         int flags = currentSelectId(Py_TYPE(self));
@@ -554,16 +611,14 @@ PyObject *getHiddenDataFromQObject(QObject *cppSelf, PyObject *self, PyObject *n
             // Note: before implementing this property handling, the meta function code
             // below created meta functions which was quite wrong.
             auto *subdict = _PepType_Lookup(Py_TYPE(self), PySideMagicName::property_methods());
-            PyObject *propName = PyDict_GetItem(subdict, name);
-            if (propName) {
+            if (PyObject *propName = PyDict_GetItem(subdict, name)) {
                 // We really have a property name and need to fetch the fget or fset function.
                 static PyObject *const _fget = Shiboken::String::createStaticString("fget");
                 static PyObject *const _fset = Shiboken::String::createStaticString("fset");
                 static PyObject *const _fdel = Shiboken::String::createStaticString("fdel");
                 static PyObject *const arr[3] = {_fget, _fset, _fdel};
-                auto prop = _PepType_Lookup(Py_TYPE(self), propName);
-                for (int idx = 0; idx < 3; ++idx) {
-                    auto *trial = arr[idx];
+                auto *prop = _PepType_Lookup(Py_TYPE(self), propName);
+                for (auto *trial : arr) {
                     auto *res = PyObject_GetAttr(prop, trial);
                     if (res) {
                         AutoDecRef elemName(PyObject_GetAttr(res, PySideMagicName::name()));
@@ -597,19 +652,16 @@ PyObject *getHiddenDataFromQObject(QObject *cppSelf, PyObject *self, PyObject *n
                 if (methMatch) {
                     if (method.methodType() == QMetaMethod::Signal) {
                         signalList.append(method);
-                    } else {
-                        PySideMetaFunction *func = MetaFunction::newObject(cppSelf, i);
-                        if (func) {
-                            PyObject *result = reinterpret_cast<PyObject *>(func);
-                            PyObject_SetAttr(self, name, result);
-                            return result;
-                        }
+                    } else if (auto *func = MetaFunction::newObject(cppSelf, i)) {
+                        auto *result = reinterpret_cast<PyObject *>(func);
+                        PyObject_SetAttr(self, name, result);
+                        return result;
                     }
                 }
             }
             if (!signalList.isEmpty()) {
-                PyObject *pySignal = reinterpret_cast<PyObject *>(
-                    Signal::newObjectFromMethod(self, signalList));
+                auto *pySignal = reinterpret_cast<PyObject *>(
+                    Signal::newObjectFromMethod(cppSelf, self, signalList));
                 PyObject_SetAttr(self, name, pySignal);
                 return pySignal;
             }
@@ -666,15 +718,44 @@ static void invalidatePtr(any_t *object)
     if (Py_IsInitialized() == 0)
         return;
 
-    Shiboken::GilState state;
-
-    SbkObject *wrapper = Shiboken::BindingManager::instance().retrieveWrapper(object);
-    if (wrapper != nullptr)
-        Shiboken::BindingManager::instance().releaseWrapper(wrapper);
+    // Check for existence before locking (fix hang when QObjects
+    // are moved to different threads).
+    auto &bindingManager = Shiboken::BindingManager::instance();
+    if (bindingManager.hasWrapper(object)) {
+        Shiboken::GilState state;
+        if (SbkObject *wrapper = bindingManager.retrieveWrapper(object))
+            bindingManager.releaseWrapper(wrapper);
+    }
 }
 
 static const char invalidatePropertyName[] = "_PySideInvalidatePtr";
 
+// PYSIDE-2749: Skip over internal QML classes and classes
+// with dynamic meta objects when looking for the best matching
+// type to avoid unnessarily triggering the lazy load mechanism
+// for classes that do not have a binding from things like eventFilter().
+static inline bool isInternalObject(const char *name)
+{
+    return std::strstr(name, "QMLTYPE") != nullptr || std::strstr(name, "QQmlPrivate") != nullptr;
+}
+
+static const QMetaObject *metaObjectCandidate(const QObject *o)
+{
+    const auto *metaObject = o->metaObject();
+    // Skip QML helper types and Python objects
+    if (hasDynamicMetaObject(o)) {
+        if (const auto *super = metaObject->superClass())
+            metaObject = super;
+    }
+    for (const auto *candidate = metaObject; candidate != nullptr; candidate = candidate->superClass()) {
+        if (!isInternalObject(candidate->className())) {
+            metaObject = candidate;
+            break;
+        }
+    }
+    return metaObject;
+}
+
 // PYSIDE-1214, when creating new wrappers for classes inheriting QObject but
 // not exposed to Python, try to find the best-matching (most-derived) Qt
 // class by walking up the meta objects.
@@ -682,19 +763,13 @@ static const char *typeName(const QObject *cppSelf)
 {
     const char *typeName = typeid(*cppSelf).name();
     if (!Shiboken::Conversions::getConverter(typeName)) {
-        for (auto metaObject = cppSelf->metaObject(); metaObject; metaObject = metaObject->superClass()) {
+        const auto *metaObject = metaObjectCandidate(cppSelf);
+        for (; metaObject != nullptr; metaObject = metaObject->superClass()) {
             const char *name = metaObject->className();
             if (Shiboken::Conversions::getConverter(name)) {
                 typeName = name;
                 break;
             }
-            // PYSIDE-2404: Did not find the name. Load the lazy classes
-            //              which have this name and try again.
-            Shiboken::Module::loadLazyClassesWithName(name);
-            if (Shiboken::Conversions::getConverter(name)) {
-                typeName = name;
-                break;
-            }
         }
     }
     return typeName;
@@ -708,15 +783,12 @@ PyTypeObject *getTypeForQObject(const QObject *cppSelf)
     if (existing != nullptr)
         return reinterpret_cast<PyObject *>(existing)->ob_type;
     // Find the best match (will return a PySide type)
-    auto *sbkObjectType = Shiboken::ObjectType::typeForTypeName(typeName(cppSelf));
-    if (sbkObjectType != nullptr)
-        return reinterpret_cast<PyTypeObject *>(sbkObjectType);
-    return nullptr;
+    return Shiboken::ObjectType::typeForTypeName(typeName(cppSelf));
 }
 
 PyObject *getWrapperForQObject(QObject *cppSelf, PyTypeObject *sbk_type)
 {
-    PyObject *pyOut = reinterpret_cast<PyObject *>(Shiboken::BindingManager::instance().retrieveWrapper(cppSelf));
+    auto *pyOut = reinterpret_cast<PyObject *>(Shiboken::BindingManager::instance().retrieveWrapper(cppSelf));
     if (pyOut) {
         Py_INCREF(pyOut);
         return pyOut;
@@ -738,7 +810,7 @@ PyObject *getWrapperForQObject(QObject *cppSelf, PyTypeObject *sbk_type)
         }
     }
 
-    pyOut = Shiboken::Object::newObject(sbk_type, cppSelf, false, false, typeName(cppSelf));
+    pyOut = Shiboken::Object::newObjectWithHeuristics(sbk_type, cppSelf, false, typeName(cppSelf));
 
     return pyOut;
 }
@@ -770,17 +842,17 @@ PyObject *qStringToPyUnicode(QStringView s)
 QString pyStringToQString(PyObject *str)
 {
     if (str == Py_None)
-        return QString();
+        return {};
 
     if (PyUnicode_Check(str) != 0)
         return pyUnicodeToQString(str);
 
     if (PyBytes_Check(str)) {
-        const char *asciiBuffer = PyBytes_AS_STRING(str);
+        const char *asciiBuffer = PyBytes_AsString(str);
         if (asciiBuffer)
             return QString::fromLatin1(asciiBuffer);
     }
-    return QString();
+    return {};
 }
 
 // PySide-1499: Provide an efficient, correct PathLike interface
@@ -789,7 +861,7 @@ QString pyPathToQString(PyObject *path)
     // For empty constructors path can be nullptr
     // fallback to an empty QString in that case.
     if (!path)
-        return QString();
+        return {};
 
     // str or bytes pass through
     if (PyUnicode_Check(path) || PyBytes_Check(path))
@@ -798,15 +870,13 @@ QString pyPathToQString(PyObject *path)
     // Let PyOS_FSPath do its work and then fix the result for Windows.
     Shiboken::AutoDecRef strPath(PyOS_FSPath(path));
     if (strPath.isNull())
-        return QString();
+        return {};
     return QDir::fromNativeSeparators(pyStringToQString(strPath));
 }
 
 bool isCompiledMethod(PyObject *callback)
 {
-    return PyObject_HasAttr(callback, PySide::PySideName::im_func())
-           && PyObject_HasAttr(callback, PySide::PySideName::im_self())
-           && PyObject_HasAttr(callback, PySide::PySideMagicName::code());
+    return Shiboken::isCompiledMethod(callback);
 }
 
 static const unsigned char qt_resource_name[] = {
@@ -856,7 +926,7 @@ bool registerInternalQtConf()
     // This will disable the internal qt.conf which points to the PySide6 subdirectory (due to the
     // subdirectory not existing anymore).
 #ifndef PYPY_VERSION
-    QString executablePath = QString::fromWCharArray(Py_GetProgramFullPath());
+    QString executablePath = QString::fromWCharArray(PyUnicode_AsWideCharString(PySys_GetObject("executable"), nullptr));
 #else
     // PYSIDE-535: FIXME: Add this function when available.
     QString executablePath = QLatin1StringView("missing Py_GetProgramFullPath");
@@ -1027,8 +1097,8 @@ static void formatPySequence(PyObject *obj, QDebug &debug)
 
 static void formatPyDict(PyObject *obj, QDebug &debug)
 {
-    PyObject *key;
-    PyObject *value;
+    PyObject *key{};
+    PyObject *value{};
     Py_ssize_t pos = 0;
     bool first = true;
     debug << '{';
@@ -1115,7 +1185,13 @@ static void formatPyObject(PyObject *obj, QDebug &debug)
         return;
     }
     if (PyType_Check(obj) == 0)
-        debug << pyTypeName(obj) << ": ";
+        debug << pyTypeName(obj);
+    const auto refs = Py_REFCNT(obj);
+    if (refs == UINT_MAX) // _Py_IMMORTAL_REFCNT
+        debug << ", immortal";
+    else
+        debug << ", refs=" << refs;
+    debug << ": ";
     formatPyObjectValue(obj, debug);
 }
 
